const config = require("config");
const dayjs = require("dayjs");
const path = require("path");
const Key = require("./Key");
const RedisClient = require("./RedisClient");
const PaymentProcessor = require("./payment_processor/PaymentProcessor");
const { CLIENT } = require("./RedisClient");
const i18next = require("i18next");

class Order {
  static get STORAGE_MUTEX_KEY_PREFIX() {
    return "order_mutex:";
  }

  static get STORAGE_KEY_PREFIX() {
    return "order:";
  }

  static get PURCHASE_TAX_AMOUNT() {
    return 0.07;
  }

  // How long is a link between order <=> key stored
  static get PURCHASE_LINK_TIME_DAYS() {
    return 31;
  }
  static get PURCHASE_LINK_ORDER_TO_KEY_PREFIX() {
    return "order_link_order_key:";
  }

  // How many minutes is a user allowed to take for finishing the payment
  static get PURCHASE_STORAGE_TIME_UNCOMPLETED_HOURS() {
    return 6;
  }

  static GET_ORDER_FILE_BASE_PATH(order_date) {
    let order_path = config.get("storage.data_path");
    order_path = path.join(order_path, process.env.NODE_ENV, "orders");
    order_path = path.join(
      order_path,
      order_date.format("YYYY"),
      order_date.format("MM")
    );

    return order_path;
  }
  /**
   * Data stored in Redis Database
   */
  #order_id;
  #order_date;
  #expires_at;
  #amount;
  #amount_refund_requested = 0;
  #amount_refunded = 0;
  #price;
  #foreign_currency_price;
  #foreign_currency;
  #order_key_charged = false;
  #payment_captured = false;
  #order_key_changes = [];

  #receipt_created = false;
  #civicrm_contribution_id;
  /** @type {PaymentProcessor} */
  #payment_processor;

  /**
   * Data populated by context
   */
  #order_path;

  constructor(
    order_id,
    order_date = null,
    amount,
    price,
    foreign_currency_price,
    foreign_currency,
    payment_processor,
    amount_refund_requested = 0,
    order_key_charged = false,
    payment_captured = false,
    amount_refunded = 0,
    receipt_created = false,
    civicrm_contribution_id,
    order_key_changes = []
  ) {
    this.#order_id = order_id;
    if (order_date !== null) {
      this.#order_date = order_date;
    } else {
      this.#order_date = dayjs.unix(this.#order_id.substr(0, 10));
    }
    this.#expires_at = dayjs().add(config.get("keys.expiration_days"), "days");


    this.#order_path = Order.GET_ORDER_FILE_BASE_PATH(this.#order_date);

    this.#amount = parseInt(amount);
    this.#price = parseFloat(price);
    if (foreign_currency_price !== undefined) {
      this.#foreign_currency_price = parseFloat(foreign_currency_price);
    }
    if (foreign_currency !== undefined) {
      this.#foreign_currency = foreign_currency;
    }

    this.#payment_processor = payment_processor;

    this.#amount_refund_requested = amount_refund_requested;
    this.#order_key_charged = order_key_charged;
    this.#payment_captured = payment_captured;
    this.#amount_refunded = amount_refunded;

    this.#receipt_created = receipt_created;

    this.#civicrm_contribution_id = civicrm_contribution_id;
    this.#order_key_changes = order_key_changes;
  }

  getOrderID() {
    return this.#order_id;
  }

  getAmount() {
    return this.#amount;
  }

  getNettoPrice() {
    let amount = this.getPrice();
    let vat = config.get("price.vat") / 100; // Vat is configured in %
    return amount / (1 + vat);
  }
  getVat() {
    return config.get("price.vat");
  }
  getVatAmount() {
    let vat = config.get("price.vat") / 100; // Vat is configured in %
    let amount = this.getPrice();
    return (amount / (1 + vat)) * vat;
  }
  getPrice() {
    return this.#price;
  }
  getForeignCurrencyPrice() {
    return this.#foreign_currency_price;
  }
  getForeignCurrency() {
    return this.#foreign_currency;
  }

  /**
   *
   * @returns {PaymentProcessor}
   */
  getPaymentProcessor() {
    return this.#payment_processor;
  }

  /**
   *
   * @param {dayjs.Dayjs} expiration
   */
  setExpiration(expiration) {
    this.#expires_at = expiration;
  }

  async captureOrder() {
    if (this.isPaymentCaptured()) {
      throw new Error("Order already captured");
    }
    return this.getWriteLock()
      .then(() => this.getPaymentProcessor().captureOrder())
      .then(() => this.createCiviCRMOrder())
      .then(() => {
        this.#payment_captured = true;
        this.#order_date = dayjs();
        return this.save();
      })
      .then(() => this.releaseWriteLock());
  }

  async chargeKey() {
    if (this.isOrderKeyCharged()) {
      throw new Error("Key already charged");
    }
    return this.getWriteLock()
      .then(() => this.getKeyFromOrderLink())
      .then((key) => Key.GET_KEY(key, true))
      .then((key) => {
        key.charge_key_order(
          this.getAmount(),
          this.getOrderID(),
          this.#expires_at
        );
        return key.save();
      })
      .then(() => {
        this.#order_key_charged = true;
        return this.save();
      })
      .then(() => this.releaseWriteLock());
  }

  async requestRefund(amount) {
    if (this.#amount_refund_requested > 0) {
      throw new Error("Refund is already requested");
    }
    let new_key = null;
    return this.getWriteLock()
      .then(() => this.getKeyFromOrderLink())
      .then((key) => Key.GET_KEY(key, true))
      .then((key) => {
        key.discharge_key(amount, this.getOrderID());
        return key.save();
      })
      .then((key) => {
        new_key = key;
        this.#amount_refund_requested = amount;
        return this.save();
      })
      .then(() => this.releaseWriteLock())
      .then(() => new_key);
  }

  /**
   * Processes a refund which was previously issued by the customer.
   * Will issue a refund from the payment processor
   *
   * @returns
   */
  async refund() {
    if (this.#amount_refunded > 0 || this.#amount_refund_requested === 0) {
      throw "No refund requested or Refund already processed";
    }

    let refund_price = parseFloat(
      (
        Math.ceil(
          (this.#amount_refund_requested / this.#amount) * this.#price * 100
        ) / 100
      ).toFixed(2)
    );

    return this.getWriteLock()
      .then(() => this.getPaymentProcessor().refundOrder(refund_price))
      .then(() => {
        this.#amount_refunded = this.#amount_refund_requested;
        return this.save();
      })
      .then(() => this.releaseWriteLock());
  }

  /**
   * Processes an external refund. I.e. when the refund was issued through Payment Gateway directly
   *
   * @param {float} price price that was refunded externally
   */
  async external_refund(price) {
    // Create amount from refunded price
    price = Math.min(this.#price, price);
    let amount = Math.floor((price / this.#price) * this.#amount);

    return this.getWriteLock()
      .then(() => this.getKeyFromOrderLink())
      .then((key) => Key.GET_KEY(key, true))
      .then((key) => {
        try {
          key.discharge_key(amount, this.getOrderID());
          return key.save();
        } catch (err) {
          throw err;
        }
      })
      .then(() => {
        this.#amount_refunded = amount;
        return this.save();
      })
      .then(() => this.releaseWriteLock());
  }

  async denyRefund() {
    if (this.#amount_refunded > 0 || this.#amount_refund_requested === 0) {
      return false;
    }
    return this.getWriteLock()
      .then(() => this.getKeyFromOrderLink())
      .then((key) => Key.GET_KEY(key, true))
      .then((key) => {
        key.charge_key_order(
          this.#amount_refund_requested,
          this.getOrderID(),
          this.#expires_at
        );
        return key.save();
      })
      .then(() => {
        this.#amount_refund_requested = 0;
        return this.save();
      })
      .then(() => this.releaseWriteLock())
      .then(() => {
        return true;
      })
      .catch((reason) => {
        console.debug(reason);
        return false;
      });
  }

  isReceiptCreated() {
    return this.#receipt_created;
  }

  getAmountRefundRequested() {
    return this.#amount_refund_requested;
  }

  isOrderKeyCharged() {
    return this.#order_key_charged;
  }
  isPaymentCaptured() {
    return this.#payment_captured;
  }
  getAmountRefunded() {
    return this.#amount_refunded;
  }

  /**
   *
   * @param {int} amount
   * @param {string} key
   * @param {PaymentProcessor} payment_processor
   * @param {i18next.TFunction} t
   * @returns
   */
  static async CREATE_NEW_ORDER(amount, key, payment_processor, t) {
    // Calculate price from amount
    let price = (amount / 300) * config.get("price.per_300");
    price = price.toFixed(2);

    // Validate that Key can be charged with another order
    /**
     * @type {Key}
     */
    let loaded_key = null;
    return Key.GET_KEY(key)
      .then((key) => {
        if (!key.isChargable()) {
          throw "Key cannot be charged";
        } else {
          loaded_key = key;
          return true;
        }
      })
      .then(() => Order.GENERATE_UNIQUE_ORDER_ID())
      .then((order_id) => {
        let new_order = new Order(
          order_id,
          null,
          amount,
          price,
          null,
          null,
          payment_processor
        );
        return new_order
          .getWriteLock()
          .then(() => new_order.getPaymentProcessor().createOrder(new_order, t)) // Create Order on PaymentProcessor side
          .then(() => new_order.createOrderLink(loaded_key.get_key()))
          .then(() => new_order.save())
          .then(() => new_order.releaseWriteLock())
          .then(() => {
            return new_order;
          });
      });
  }

  /**
   *
   * @param {Number} order_id
   * @returns {Promise<Order>}
   */
  static async LOAD_ORDER_FROM_ID(order_id) {
    return new Promise((resolve, reject) => {
      __redis_client
        .get(Order.STORAGE_KEY_PREFIX + order_id)
        .then((order_data) => {
          let order_date = dayjs.unix(order_id.substr(0, 10));
          if (order_data === null) {
            let order_file = path.join(
              Order.GET_ORDER_FILE_BASE_PATH(order_date),
              order_id.toString() + ".json"
            );
            let fs = require("fs");
            if (fs.existsSync(order_file)) {
              order_data = JSON.parse(fs.readFileSync(order_file));
              if ("order_date" in order_data && order_data.order_date) {
                order_date = dayjs(order_data.order_date);
              }
            } else {
              throw "Could not find Order in our database!";
            }
          } else {
            order_data = JSON.parse(order_data);
          }
          let loaded_order = new Order(
            order_data.order_id,
            order_date,
            order_data.amount,
            order_data.price,
            order_data.foreign_currency_price,
            order_data.foreign_currency,
            PaymentProcessor.LOAD_PAYMENT_PROCESSOR(
              order_data.payment_processor
            ),
            order_data.amount_refund_requested,
            order_data.order_key_charged,
            order_data.payment_captured,
            order_data.amount_refunded,
            order_data.receipt_created,
            order_data.civicrm_contribution_id,
            order_data.order_key_changes
          );
          resolve(loaded_order);
        })
        .catch((reason) => {
          reject(reason);
        });
    });
  }

  async save(redis_expiration = null) {
    let stored_data = {
      order_id: this.#order_id,
      order_date: this.#order_date.unix(),
      expires_at: this.#expires_at,
      amount: this.#amount,
      price: this.#price,
      foreign_currency_price: this.#foreign_currency_price,
      foreign_currency: this.#foreign_currency,
      amount_refund_requested: this.#amount_refund_requested,
      order_key_charged: this.#order_key_charged,
      payment_captured: this.#payment_captured,
      amount_refunded: this.#amount_refunded,
      receipt_created: this.#receipt_created,
      payment_processor: this.#payment_processor.serialize(),
      civicrm_contribution_id: this.#civicrm_contribution_id,
      order_key_changes: this.#order_key_changes,
    };
    /**
     * Completed Orders will be stored in Filesystem
     * Uncompleted Orders will be stored in Redis
     */
    if (redis_expiration === null) {
      if (this.isPaymentCaptured()) {
        redis_expiration = dayjs().add(Order.PURCHASE_LINK_TIME_DAYS, "day");
      } else {
        redis_expiration = dayjs().add(
          Order.PURCHASE_STORAGE_TIME_UNCOMPLETED_HOURS,
          "hour"
        );
      }
    } else {
      let min_expiration = dayjs().add(
        Order.PURCHASE_STORAGE_TIME_UNCOMPLETED_HOURS,
        "hour"
      );
      if (this.isPaymentCaptured()) {
        min_expiration = dayjs().add(Order.PURCHASE_LINK_TIME_DAYS, "day");
      }
      if (min_expiration.isAfter(redis_expiration)) {
        redis_expiration = min_expiration;
      }
    }

    let redis_key = Order.STORAGE_KEY_PREFIX + this.#order_id;
    if (this.isPaymentCaptured()) {
      let fs = require("fs");
      let order_file = path.join(
        this.#order_path,
        this.#order_id.toString() + ".json"
      );
      // Create directory if it does not exist
      if (!fs.existsSync(path.dirname(order_file))) {
        fs.mkdirSync(path.dirname(order_file), { recursive: true });
      }
      return __redis_client
        .del(redis_key)
        .then(() => this.updateOrderLinkTTL(redis_expiration))
        .then(() =>
          fs.writeFileSync(order_file, JSON.stringify(stored_data, null, 4))
        );
    } else {
      // Store Order in Redis
      return __redis_client
        .pipeline()
        .set(redis_key, JSON.stringify(stored_data))
        .expireat(redis_key, redis_expiration.unix())
        .exec()
        .then(() => this.updateOrderLinkTTL(redis_expiration));
    }
  }

  async delete() {
    let redis_key = Order.STORAGE_KEY_PREFIX + this.#order_id;
    if (!this.isPaymentCaptured()) {
      return this.deleteOrderLink()
        .then(() => __redis_client.del(redis_key))
        .then(async (deleted_keys) => {
          if (deleted_keys > 0) {
            return true;
          } else {
            return false;
          }
        })
        .catch(async () => {
          return false;
        });
    } else {
      return false;
    }
  }

  async getWriteLock() {
    let write_lock_key = Order.STORAGE_MUTEX_KEY_PREFIX + this.#order_id;

    return new Promise(async (resolve, reject) => {
      let timestart = dayjs();
      let write_lock = 0;
      do {
        write_lock = await __redis_client
          .pipeline()
          .setnx(write_lock_key, 1)
          .expiretime(write_lock_key)
          .expire(write_lock_key, 15)
          .exec();
        let expire_at = write_lock[1][1];
        write_lock = write_lock[0][1];
        if (write_lock === 1) {
          resolve(true);
          break;
        } else if (dayjs().diff(timestart, "second") >= 15) {
          if (expire_at >= 0) {
            await __redis_client.expireat(write_lock_key, expire_at);
          }
          reject("Timed out waiting for write lock for Order");
          break;
        } else {
          if (expire_at >= 0) {
            await __redis_client.expireat(write_lock_key, expire_at);
          }
          await new Promise((resolve) => setTimeout(resolve, 1000));
        }
      } while (true);
    });
  }

  async releaseWriteLock() {
    let write_lock_key = Order.STORAGE_MUTEX_KEY_PREFIX + this.#order_id;
    return __redis_client.del(write_lock_key);
  }

  addOrderKeyChanges(changes) {
    this.#order_key_changes.push(changes);
  }

  /**
   * Creates a link between Order and Key that will automatically expire
   * @param {string} key
   */
  async createOrderLink(key) {
    let expiration = dayjs().add(Order.PURCHASE_LINK_TIME_DAYS, "day");
    let redis_key_order_key =
      Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;
    return __redis_client
      .pipeline()
      .set(redis_key_order_key, key)
      .expireat(redis_key_order_key, expiration.unix())
      .exec();
  }

  async deleteOrderLink() {
    let redis_key_order_key =
      Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;
    return this.getKeyFromOrderLink().then((key) => {
      return __redis_client.pipeline().del(redis_key_order_key).exec();
    });
  }

  async createCiviCRMOrder() {
    if (this.getPaymentProcessor().serialize().processor_name === "Manual") {
      // Manual Orders will not get stored in CiviCRM
      return;
    }

    let civicrm_data = config.get("app.civicrm");
    if (!civicrm_data.enabled) {
      return;
    }
    return fetch(civicrm_data.url + "/Contribution/create", {
      method: "post",
      headers: {
        "Content-Type": "application/x-www-form-urlencoded",
        "X-Civi-Auth": `Bearer ${civicrm_data.api_key}`,
      },
      body:
        "params=" +
        encodeURIComponent(
          JSON.stringify({
            values: {
              contact_id: civicrm_data.user_id,
              financial_type_id: 5,
              receive_date: this.#order_date.format("YYYY-MM-DD HH:mm:ss"),
              total_amount: this.getPrice(),
              is_test: process.env.NODE_ENV === "development" ? true : false,
              is_pay_later: false,
              tax_amount: this.getVatAmount(),
              is_template: false,
            },
          })
        ),
    })
      .then((response) => {
        if (response.status !== 200) {
          throw "Error Creating Civicrm Contribution";
        } else {
          return response.json();
        }
      })
      .then((response_json) => {
        this.#civicrm_contribution_id = response_json.values[0].id;
      });
  }

  getOrderDate() {
    return this.#order_date;
  }

  /**
   *
   * @param {number} price
   * @param {number} foreign_price
   * @param {string} foreign_currency
   */
  setPrice(price, foreign_price = 0, foreign_currency = "EUR") {
    if (
      config.get("price.allowed_currencies").indexOf(foreign_currency) === -1
    ) {
      throw new Error(`Currency ${foreign_currency} not allowed`);
    }
    if (this.isPaymentCaptured()) {
      throw new Error("Cannot change details of processed Order");
    }
    this.#price = price;
    this.#foreign_currency_price = foreign_price;
    this.#foreign_currency = foreign_currency;

    // Calculate new amount from price
    this.#amount = parseInt(
      Math.ceil((price / config.get("price.per_300")) * 300)
    );
  }

  async attachReceipt(data, invoice_id) {
    let civicrm_data = config.get("app.civicrm");
    if (!civicrm_data.enabled) {
      return;
    }

    return this.getWriteLock()
      .then(() =>
        fetch(civicrm_data.url + "/Contribution/addReceipt", {
          method: "post",
          headers: {
            "Content-Type": "application/x-www-form-urlencoded",
            "X-Civi-Auth": `Bearer ${civicrm_data.api_key}`,
          },
          body:
            "params=" +
            encodeURIComponent(
              JSON.stringify({
                receipt: data,
                invoice_id,
                id: this.#civicrm_contribution_id,
              })
            ),
        })
      )
      .then((response) => {
        if (response.status !== 200) {
          throw "Error Creating Civicrm Contribution";
        } else {
          return response.json();
        }
      })
      .then((response_json) => {
        if (response_json["count"] !== 1) {
          throw JSON.stringify(response_json, null, 4);
        } else {
          this.#receipt_created = true;
          return response_json;
        }
      })
      .then(() => this.save())
      .then(() => this.releaseWriteLock());
  }

  async getReceipt() {
    let civicrm_data = config.get("app.civicrm");
    if (!civicrm_data.enabled) {
      return Promise.reject("No Civicrm Config supplied");
    }
    return fetch(civicrm_data.url + "/Contribution/getReceipt", {
      method: "post",
      headers: {
        "Content-Type": "application/x-www-form-urlencoded",
        "X-Civi-Auth": `Bearer ${civicrm_data.api_key}`,
      },
      body:
        "params=" +
        encodeURIComponent(
          JSON.stringify({
            id: this.#civicrm_contribution_id,
          })
        ),
    })
      .then((result) => result.json())
      .then((result_json) => {
        if (
          typeof result_json.error_code !== "undefined" &&
          result_json.error_code !== 0
        ) {
          throw result_json;
        } else {
          return result_json;
        }
      });
  }

  /**
   *
   * @param {dayjs.Dayjs} redis_expiration
   * @returns
   */
  async updateOrderLinkTTL(redis_expiration) {
    let redis_key = Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;

    return __redis_client.expireat(redis_key, redis_expiration.unix());
  }

  /**
   * Tries to get a key for this order. If the order link already expired this function rejects the promise
   */
  async getKeyFromOrderLink() {
    let redis_key = Order.PURCHASE_LINK_ORDER_TO_KEY_PREFIX + this.#order_id;
    return __redis_client.get(redis_key).then(async (result) => {
      if (result) {
        return result;
      } else {
        throw "Order Link does not exist";
      }
    });
  }

  static async GENERATE_UNIQUE_ORDER_ID() {
    // Generate Order ID => time in seconds since 1.1.1970 and and add a mutex to it to allow multiple order ids per second
    let order_id = null;
    let order_base = Math.round(new Date().getTime() / 1000);
    let order_mutex = Math.floor(Math.random() * 10000);
    do {
      // make sure this order_id is not already registered
      order_id = "" + order_base + order_mutex;
      let order_lock = await __redis_client.setnx(order_id, true);
      if (order_lock === 1) {
        await __redis_client.expire(order_id, 5);
      } else {
        order_mutex++;
        order_id = null;
      }
    } while (order_id === null);
    return order_id;
  }
}

module.exports = Order;
