const Order = require("../Order");
const PaymentProcessor = require("./PaymentProcessor");
const i18next = require("i18next");
const config = require("config");

class Micropayment extends PaymentProcessor {
  static get NAME() { return "Micropayment" };
  /**
   * @var {string[]}
   */
  #supported_services = ["paysafecard"];

  /**
   * @var {string}
   */
  #service;

  /**
   * @var {URL}
   */
  #capture_url;

  /**
   *
   * @param {string} service
   */
  constructor(service = "paysafecard", capture_url = null) {
    super();

    if (this.#supported_services.indexOf(service) === -1) {
      throw new Error(`Payment Service ${service} is not supported.`);
    }
    this.#service = service;
    this.#capture_url = capture_url;
  }

  /**
   * Generates a URL for the customer where he is redirected to
   * to complete Payment.
   *
   * @param {Order} order
   * @param {i18next.TFunction} t
   *
   * @returns {Promise<URL>}
   */
  async createOrder(order, t) {
    let payment_window_url = new URL(
      `https://${this.#service}.micropayment.de/${this.#service}/event`
    );
    let searchParams = new URLSearchParams({
      project: config.get("payments.micropayment.project"),
      amount: order.getPrice() * 100,
      title: t("product.name", { ns: "order" }),
      mp_user_id: order.getOrderID(),
      producttype: "quantity",
      paytext: t("product.name", { ns: "order", count: order.getAmount() }),
      testmode: process.env.NODE_ENV === "development" ? "1" : "0",
      orderid: order.getOrderID(),
      vatinfo: "1",
    });
    payment_window_url.search = searchParams.toString();
    this.#capture_url = payment_window_url;
    return Promise.resolve(this.#capture_url);
  }

  /**
   * Generates a URL for the customer to be redirected to after Payment
   *
   * @returns {Promise<URL>}
   */
  async captureOrder() {
    return Promise.resolve();
  }

  /**
 * 
 * @returns {boolean}
 */
  isRefundSupported() {
    return true;
  }

  async refundOrder() {
    throw new Error("Manual Payments cannot be refunded");
  }

  static async VERIFY_WEBHOOK(query) {
    // Verify that secretkey is correct
    if (
      !("secretfield" in query) ||
      query.secretfield !== config.get("payments.micropayment.secretfield")
    ) {
      return Promise.reject("Secretfield does not match");
    }
    if (
      !("testmode" in query) ||
      (query.testmode === "1" && process.env.NODE_ENV !== "development")
    ) {
      return Promise.reject("Testmode Payment not accepted in production");
    }
    if (!("currency" in query) || query.currency !== "EUR") {
      return Promise.reject("Can only receive Payments in EUR")
    }
    return Promise.resolve();
  }

  /**
   * Turns class object into serializable data to store in redis db
   *
   * @returns {Object}
   */
  serialize() {
    return {
      processor_name: "Micropayment",
      service: this.#service,
      capture_url: this.#capture_url,
    };
  }

  /**
   * Turns class object into serializable data to store in redis db
   *
   * @returns {PaymentProcessor}
   */
  static DESERIALIZE(data) {
    return new Micropayment(data.service, data.capture_url);
  }
}

module.exports = Micropayment;
